/*
  This file is part of TALER
  Copyright (C) 2014-2026 Taler Systems SA

  TALER is free software; you can redistribute it and/or modify it under the
  terms of the GNU Lesser General Public License as published by the Free Software
  Foundation; either version 2.1, or (at your option) any later version.

  TALER is distributed in the hope that it will be useful, but WITHOUT ANY
  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR
  A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public License along with
  TALER; see the file COPYING.LGPL.  If not, see
  <http://www.gnu.org/licenses/>
*/
/**
 * @file src/lib/merchant_api_get-private-orders.c
 * @brief Implementation of the GET /private/orders request
 * @author Christian Grothoff
 */
#include "platform.h"
#include <curl/curl.h>
#include <jansson.h>
#include <microhttpd.h> /* just for HTTP status codes */
#include <gnunet/gnunet_util_lib.h>
#include <gnunet/gnunet_curl_lib.h>
#include <taler/merchant/get-private-orders.h>
#include "merchant_api_curl_defaults.h"
#include <taler/taler_json_lib.h>


/**
 * Maximum number of orders we return.
 */
#define MAX_ORDERS 1024


/**
 * Handle for a GET /private/orders operation.
 */
struct TALER_MERCHANT_GetPrivateOrdersHandle
{
  /**
   * Base URL of the merchant backend.
   */
  char *base_url;

  /**
   * The full URL for this request.
   */
  char *url;

  /**
   * Handle for the request.
   */
  struct GNUNET_CURL_Job *job;

  /**
   * Function to call with the result.
   */
  TALER_MERCHANT_GetPrivateOrdersCallback cb;

  /**
   * Closure for @a cb.
   */
  TALER_MERCHANT_GET_PRIVATE_ORDERS_RESULT_CLOSURE *cb_cls;

  /**
   * Reference to the execution context.
   */
  struct GNUNET_CURL_Context *ctx;

  /**
   * Paid filter.
   */
  enum TALER_EXCHANGE_YesNoAll paid;

  /**
   * Refunded filter.
   */
  enum TALER_EXCHANGE_YesNoAll refunded;

  /**
   * Wired filter.
   */
  enum TALER_EXCHANGE_YesNoAll wired;

  /**
   * Date threshold for filtering.
   */
  struct GNUNET_TIME_Timestamp date;

  /**
   * Starting row for pagination.
   */
  uint64_t offset;

  /**
   * Limit on number of results.
   */
  int64_t limit;

  /**
   * Long polling timeout.
   */
  struct GNUNET_TIME_Relative timeout;

  /**
   * Session ID filter, or NULL.
   */
  char *session_id;

  /**
   * Fulfillment URL filter, or NULL.
   */
  char *fulfillment_url;

  /**
   * Summary text filter, or NULL.
   */
  char *summary_filter;

  /**
   * Maximum age filter.
   * @since protocol v27.
   */
  struct GNUNET_TIME_Relative max_age;

  /**
   * True if offset was explicitly set.
   */
  bool have_offset;

  /**
   * True if date was explicitly set.
   */
  bool have_date;

  /**
   * True if @e max_age was set.
   */
  bool have_max_age;
};


/**
 * Parse order information from @a ia.
 *
 * @param ia JSON array (or NULL!) with order data
 * @param[in,out] ogr response to fill
 * @param oph operation handle
 * @return #GNUNET_OK on success
 */
static enum GNUNET_GenericReturnValue
parse_orders (const json_t *ia,
              struct TALER_MERCHANT_GetPrivateOrdersResponse *ogr,
              struct TALER_MERCHANT_GetPrivateOrdersHandle *oph)
{
  unsigned int oes_len = (unsigned int) json_array_size (ia);

  if ( (json_array_size (ia) != (size_t) oes_len) ||
       (oes_len > MAX_ORDERS) )
  {
    GNUNET_break (0);
    return GNUNET_SYSERR;
  }
  {
    struct TALER_MERCHANT_GetPrivateOrdersOrderEntry oes[
      GNUNET_NZL (oes_len)];
    size_t index;
    json_t *value;

    memset (oes,
            0,
            sizeof (oes));
    json_array_foreach (ia, index, value) {
      struct TALER_MERCHANT_GetPrivateOrdersOrderEntry *ie = &oes[index];
      struct GNUNET_JSON_Specification spec[] = {
        TALER_JSON_spec_slug ("order_id",
                              &ie->order_id),
        GNUNET_JSON_spec_timestamp ("timestamp",
                                    &ie->timestamp),
        GNUNET_JSON_spec_uint64 ("row_id",
                                 &ie->order_serial),
        TALER_JSON_spec_amount_any ("amount",
                                    &ie->amount),
        GNUNET_JSON_spec_mark_optional (
          TALER_JSON_spec_amount_any ("refund_amount",
                                      &ie->refund_amount),
          NULL),
        GNUNET_JSON_spec_mark_optional (
          TALER_JSON_spec_amount_any ("pending_refund_amount",
                                      &ie->pending_refund_amount),
          NULL),
        GNUNET_JSON_spec_string ("summary",
                                 &ie->summary),
        GNUNET_JSON_spec_bool ("refundable",
                               &ie->refundable),
        GNUNET_JSON_spec_bool ("paid",
                               &ie->paid),
        GNUNET_JSON_spec_end ()
      };

      if (GNUNET_OK !=
          GNUNET_JSON_parse (value,
                             spec,
                             NULL, NULL))
      {
        GNUNET_break_op (0);
        return GNUNET_SYSERR;
      }
    }
    ogr->details.ok.orders_length = oes_len;
    ogr->details.ok.orders = oes;
    oph->cb (oph->cb_cls,
             ogr);
    oph->cb = NULL;
  }
  return GNUNET_OK;
}


/**
 * Function called when we're done processing the
 * HTTP GET /private/orders request.
 *
 * @param cls the `struct TALER_MERCHANT_GetPrivateOrdersHandle`
 * @param response_code HTTP response code, 0 on error
 * @param response response body, NULL if not in JSON
 */
static void
handle_get_orders_finished (void *cls,
                            long response_code,
                            const void *response)
{
  struct TALER_MERCHANT_GetPrivateOrdersHandle *oph = cls;
  const json_t *json = response;
  struct TALER_MERCHANT_GetPrivateOrdersResponse ogr = {
    .hr.http_status = (unsigned int) response_code,
    .hr.reply = json
  };

  oph->job = NULL;
  GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
              "Got /private/orders response with status code %u\n",
              (unsigned int) response_code);
  switch (response_code)
  {
  case MHD_HTTP_OK:
    {
      const json_t *orders;
      struct GNUNET_JSON_Specification spec[] = {
        GNUNET_JSON_spec_array_const ("orders",
                                      &orders),
        GNUNET_JSON_spec_end ()
      };

      if (GNUNET_OK !=
          GNUNET_JSON_parse (json,
                             spec,
                             NULL, NULL))
      {
        ogr.hr.http_status = 0;
        ogr.hr.ec = TALER_EC_GENERIC_INVALID_RESPONSE;
        break;
      }
      if (GNUNET_OK ==
          parse_orders (orders,
                        &ogr,
                        oph))
      {
        TALER_MERCHANT_get_private_orders_cancel (oph);
        return;
      }
      ogr.hr.http_status = 0;
      ogr.hr.ec = TALER_EC_GENERIC_INVALID_RESPONSE;
      break;
    }
  case MHD_HTTP_UNAUTHORIZED:
    ogr.hr.ec = TALER_JSON_get_error_code (json);
    ogr.hr.hint = TALER_JSON_get_error_hint (json);
    break;
  case MHD_HTTP_NOT_FOUND:
    ogr.hr.ec = TALER_JSON_get_error_code (json);
    ogr.hr.hint = TALER_JSON_get_error_hint (json);
    break;
  default:
    ogr.hr.ec = TALER_JSON_get_error_code (json);
    ogr.hr.hint = TALER_JSON_get_error_hint (json);
    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
                "Unexpected response code %u/%d\n",
                (unsigned int) response_code,
                (int) ogr.hr.ec);
    break;
  }
  oph->cb (oph->cb_cls,
           &ogr);
  TALER_MERCHANT_get_private_orders_cancel (oph);
}


struct TALER_MERCHANT_GetPrivateOrdersHandle *
TALER_MERCHANT_get_private_orders_create (
  struct GNUNET_CURL_Context *ctx,
  const char *url)
{
  struct TALER_MERCHANT_GetPrivateOrdersHandle *oph;

  oph = GNUNET_new (struct TALER_MERCHANT_GetPrivateOrdersHandle);
  oph->ctx = ctx;
  oph->base_url = GNUNET_strdup (url);
  oph->paid = TALER_EXCHANGE_YNA_ALL;
  oph->refunded = TALER_EXCHANGE_YNA_ALL;
  oph->wired = TALER_EXCHANGE_YNA_ALL;
  oph->date = GNUNET_TIME_UNIT_FOREVER_TS;
  oph->offset = UINT64_MAX;
  oph->limit = -20;
  return oph;
}


enum GNUNET_GenericReturnValue
TALER_MERCHANT_get_private_orders_set_options_ (
  struct TALER_MERCHANT_GetPrivateOrdersHandle *oph,
  unsigned int num_options,
  const struct TALER_MERCHANT_GetPrivateOrdersOptionValue *options)
{
  for (unsigned int i = 0; i < num_options; i++)
  {
    const struct TALER_MERCHANT_GetPrivateOrdersOptionValue *opt =
      &options[i];

    switch (opt->option)
    {
    case TALER_MERCHANT_GET_PRIVATE_ORDERS_OPTION_END:
      return GNUNET_OK;
    case TALER_MERCHANT_GET_PRIVATE_ORDERS_OPTION_PAID:
      oph->paid = opt->details.paid;
      break;
    case TALER_MERCHANT_GET_PRIVATE_ORDERS_OPTION_REFUNDED:
      oph->refunded = opt->details.refunded;
      break;
    case TALER_MERCHANT_GET_PRIVATE_ORDERS_OPTION_WIRED:
      oph->wired = opt->details.wired;
      break;
    case TALER_MERCHANT_GET_PRIVATE_ORDERS_OPTION_LIMIT:
      oph->limit = opt->details.limit;
      break;
    case TALER_MERCHANT_GET_PRIVATE_ORDERS_OPTION_OFFSET:
      oph->offset = opt->details.offset;
      oph->have_offset = true;
      break;
    case TALER_MERCHANT_GET_PRIVATE_ORDERS_OPTION_DATE:
      oph->date = opt->details.date;
      oph->have_date = true;
      break;
    case TALER_MERCHANT_GET_PRIVATE_ORDERS_OPTION_TIMEOUT:
      oph->timeout = opt->details.timeout;
      break;
    case TALER_MERCHANT_GET_PRIVATE_ORDERS_OPTION_SESSION_ID:
      GNUNET_free (oph->session_id);
      if (NULL != opt->details.session_id)
        oph->session_id = GNUNET_strdup (opt->details.session_id);
      break;
    case TALER_MERCHANT_GET_PRIVATE_ORDERS_OPTION_FULFILLMENT_URL:
      GNUNET_free (oph->fulfillment_url);
      if (NULL != opt->details.fulfillment_url)
        oph->fulfillment_url = GNUNET_strdup (opt->details.fulfillment_url);
      break;
    case TALER_MERCHANT_GET_PRIVATE_ORDERS_OPTION_SUMMARY_FILTER:
      GNUNET_free (oph->summary_filter);
      if (NULL != opt->details.summary_filter)
        oph->summary_filter = GNUNET_strdup (opt->details.summary_filter);
      break;
    case TALER_MERCHANT_GET_PRIVATE_ORDERS_OPTION_MAX_AGE:
      oph->max_age = opt->details.max_age;
      oph->have_max_age = true;
      break;
    default:
      GNUNET_break (0);
      return GNUNET_NO;
    }
  }
  return GNUNET_OK;
}


enum TALER_ErrorCode
TALER_MERCHANT_get_private_orders_start (
  struct TALER_MERCHANT_GetPrivateOrdersHandle *oph,
  TALER_MERCHANT_GetPrivateOrdersCallback cb,
  TALER_MERCHANT_GET_PRIVATE_ORDERS_RESULT_CLOSURE *cb_cls)
{
  CURL *eh;
  unsigned int tms;

  oph->cb = cb;
  oph->cb_cls = cb_cls;
  tms = (unsigned int) (oph->timeout.rel_value_us
                        / GNUNET_TIME_UNIT_MILLISECONDS.rel_value_us);
  {
    char dstr[30];
    bool have_date;
    bool have_srow;
    char cbuf[30];
    char dbuf[30];
    char tbuf[30];
    char mabuf[30];

    GNUNET_snprintf (tbuf,
                     sizeof (tbuf),
                     "%u",
                     tms);
    if (oph->have_max_age)
      GNUNET_snprintf (mabuf,
                       sizeof (mabuf),
                       "%llu",
                       (unsigned long long)
                       (oph->max_age.rel_value_us
                        / GNUNET_TIME_UNIT_SECONDS.rel_value_us));
    GNUNET_snprintf (dbuf,
                     sizeof (dbuf),
                     "%lld",
                     (long long) oph->limit);
    GNUNET_snprintf (cbuf,
                     sizeof (cbuf),
                     "%llu",
                     (unsigned long long) oph->offset);
    GNUNET_snprintf (dstr,
                     sizeof (dstr),
                     "%llu",
                     (unsigned long long) GNUNET_TIME_timestamp_to_s (
                       oph->date));
    if (oph->limit > 0)
    {
      have_date = oph->have_date
                  && ! GNUNET_TIME_absolute_is_zero (oph->date.abs_time);
      have_srow = oph->have_offset
                  && (0 != oph->offset);
    }
    else
    {
      have_date = oph->have_date
                  && ! GNUNET_TIME_absolute_is_never (oph->date.abs_time);
      have_srow = oph->have_offset
                  && (UINT64_MAX != oph->offset);
    }
    oph->url = TALER_url_join (oph->base_url,
                               "private/orders",
                               "paid",
                               (TALER_EXCHANGE_YNA_ALL != oph->paid)
                               ? TALER_yna_to_string (oph->paid)
                               : NULL,
                               "refunded",
                               (TALER_EXCHANGE_YNA_ALL != oph->refunded)
                               ? TALER_yna_to_string (oph->refunded)
                               : NULL,
                               "wired",
                               (TALER_EXCHANGE_YNA_ALL != oph->wired)
                               ? TALER_yna_to_string (oph->wired)
                               : NULL,
                               "date_s",
                               have_date
                               ? dstr
                               : NULL,
                               "offset",
                               have_srow
                               ? cbuf
                               : NULL,
                               "limit",
                               (-20 != oph->limit)
                               ? dbuf
                               : NULL,
                               "timeout_ms",
                               (0 != tms)
                               ? tbuf
                               : NULL,
                               "session_id",
                               oph->session_id,
                               "fulfillment_url",
                               oph->fulfillment_url,
                               "summary_filter",
                               oph->summary_filter,
                               "max_age_s",
                               oph->have_max_age
                               ? mabuf
                               : NULL,
                               NULL);
  }
  if (NULL == oph->url)
    return TALER_EC_GENERIC_CONFIGURATION_INVALID;
  eh = TALER_MERCHANT_curl_easy_get_ (oph->url);
  if (NULL == eh)
    return TALER_EC_GENERIC_CONFIGURATION_INVALID;
  if (0 != tms)
  {
    GNUNET_break (CURLE_OK ==
                  curl_easy_setopt (eh,
                                    CURLOPT_TIMEOUT_MS,
                                    (long) (tms + 100L)));
  }
  oph->job = GNUNET_CURL_job_add (oph->ctx,
                                  eh,
                                  &handle_get_orders_finished,
                                  oph);
  if (NULL == oph->job)
    return TALER_EC_GENERIC_INTERNAL_INVARIANT_FAILURE;
  return TALER_EC_NONE;
}


void
TALER_MERCHANT_get_private_orders_cancel (
  struct TALER_MERCHANT_GetPrivateOrdersHandle *oph)
{
  if (NULL != oph->job)
  {
    GNUNET_CURL_job_cancel (oph->job);
    oph->job = NULL;
  }
  GNUNET_free (oph->url);
  GNUNET_free (oph->session_id);
  GNUNET_free (oph->fulfillment_url);
  GNUNET_free (oph->summary_filter);
  GNUNET_free (oph->base_url);
  GNUNET_free (oph);
}


/* end of merchant_api_get-private-orders-new.c */
